@@ -12,24 +12,22 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 //
+// Copyright 2005-2010 Google, Inc.
 // Author: riley@google.com (Michael Riley)
+// Modified: jpr@google.com (Jake Ratkiewicz) to use FstClass
 //
 // \file
 // Inverts a transduction.
 //
 
-#include "./invert-main.h"
-
-namespace fst {
-
-// Register templated main for common arcs types.
-REGISTER_FST_MAIN(InvertMain, StdArc);
-REGISTER_FST_MAIN(InvertMain, LogArc);
-
-}  // namespace fst
-
+#include <fst/script/invert.h>
 
 int main(int argc, char **argv) {
+  namespace s = fst::script;
+  using fst::script::FstClass;
+  using fst::script::MutableFstClass;
+  using fst::script::VectorFstClass;
+
   string usage = "Inverts a transduction.\n\n  Usage: ";
   usage += argv[0];
   usage += " [in.fst [out.fst]]\n";
@@ -41,6 +39,22 @@ int main(int argc, char **argv) {
     return 1;
   }
 
-  // Invokes InvertMain<Arc> where arc type is determined from argv[1].
-  return CALL_FST_MAIN(InvertMain, argc, argv);
+  string in_name = (argc > 1 && strcmp(argv[1], "-") != 0) ? argv[1] : "";
+  string out_name = argc > 2 ? argv[2] : "";
+
+  FstClass *ifst = FstClass::Read(in_name);
+  if (!ifst) return 1;
+
+  MutableFstClass *ofst = 0;
+  if (ifst->Properties(fst::kMutable, false)) {
+    ofst = static_cast<MutableFstClass *>(ifst);
+  } else {
+    ofst = new VectorFstClass(*ifst);
+    delete ifst;
+  }
+
+  s::Invert(ofst);
+  ofst->Write(out_name);
+
+  return 0;
 }
