@@ -12,24 +12,22 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 //
+// Copyright 2005-2010 Google, Inc.
 // Author: riley@google.com (Michael Riley)
+// Modified: jpr@google.com (Jake Ratkiewicz) to use FstClass
 //
 // \file
 // Topologically sorts an FST.
 //
 
-#include "./topsort-main.h"
-
-namespace fst {
-
-// Register templated main for common arcs types.
-REGISTER_FST_MAIN(TopSortMain, StdArc);
-REGISTER_FST_MAIN(TopSortMain, LogArc);
-
-}  // namespace fst
-
+#include <fst/script/topsort.h>
 
 int main(int argc, char **argv) {
+  namespace s = fst::script;
+  using fst::script::FstClass;
+  using fst::script::MutableFstClass;
+  using fst::script::VectorFstClass;
+
   string usage = "Topologically sorts an FST.\n\n  Usage: ";
   usage += argv[0];
   usage += " [in.fst [out.fst]]\n";
@@ -41,6 +39,24 @@ int main(int argc, char **argv) {
     return 1;
   }
 
-  // Invokes TopSortMain<Arc> where arc type is determined from argv[1].
-  return CALL_FST_MAIN(TopSortMain, argc, argv);
+  string in_fname = argc > 1 && strcmp(argv[1], "-") != 0 ? argv[1] : "";
+  string out_fname = argc > 2 ? argv[2] : "";
+
+  FstClass *ifst = FstClass::Read(in_fname);
+  if (!ifst) return 1;
+
+  MutableFstClass *ofst = 0;
+  if (ifst->Properties(fst::kMutable, false)) {
+    ofst = static_cast<MutableFstClass *>(ifst);
+  } else {
+    ofst = new VectorFstClass(*ifst);
+    delete ifst;
+  }
+
+  bool acyclic = TopSort(ofst);
+  if (!acyclic)
+    LOG(WARNING) << argv[0] << ": Input FST is cyclic";
+  ofst->Write(out_fname);
+
+  return 0;
 }
