@@ -12,24 +12,26 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 //
+// Copyright 2005-2010 Google, Inc.
 // Author: riley@google.com (Michael Riley)
+// Modified: jpr@google.com (Jake Ratkiewicz) - to use FstClass
 //
 // \file
 // Creates the union of two FSTs.
 //
 
-#include "./union-main.h"
-
-namespace fst {
-
-// Register templated main for common arcs types.
-REGISTER_FST_MAIN(UnionMain, StdArc);
-REGISTER_FST_MAIN(UnionMain, LogArc);
-
-}  // namespace fst
+#include <string>
 
+#include <fst/script/union.h>
+#include <iostream>
+#include <fstream>
 
 int main(int argc, char **argv) {
+  using fst::script::FstClass;
+  using fst::script::MutableFstClass;
+  using fst::script::VectorFstClass;
+  using fst::script::Union;
+
   string usage = "Creates the union of two FSTs.\n\n  Usage: ";
   usage += argv[0];
   usage += " in1.fst in2.fst [out.fst]\n";
@@ -41,6 +43,36 @@ int main(int argc, char **argv) {
     return 1;
   }
 
-  // Invokes UnionMain<Arc> where arc type is determined from argv[1].
-  return CALL_FST_MAIN(UnionMain, argc, argv);
+  string in1_name = strcmp(argv[1], "-") != 0 ? argv[1] : "";
+  string in2_name = strcmp(argv[2], "-") != 0 ? argv[2] : "";
+  string out_name = argc > 3 ? argv[3] : "";
+
+  if (in1_name == "" && in2_name == "") {
+    LOG(ERROR) << argv[0]
+               << ": Can't use standard i/o for both inputs.";
+    return 1;
+  }
+
+  FstClass *ifst1 = FstClass::Read(in1_name);
+  if (!ifst1) {
+    return 0;
+  }
+
+  MutableFstClass *ofst = 0;
+  if (ifst1->Properties(fst::kMutable, false)) {
+    ofst = static_cast<MutableFstClass *>(ifst1);
+  } else {
+    ofst = new VectorFstClass(*ifst1);
+    delete ifst1;
+  }
+
+  FstClass *ifst2 = FstClass::Read(in2_name);
+  if (!ifst2) {
+    return 1;
+  }
+
+  Union(ofst, *ifst2);
+  ofst->Write(out_name);
+
+  return 0;
 }
